//
// Copyright 2002, 2006, Lowell Boggs Jr.
//
// This file or directory, containing source code for a computer program,
// is Copyrighted by Lowell Boggs, Jr.  987 Regency Drive, Lewisville
// TX (USA), 75067.  You may use, copy, modify, and distribute this
// source file without charge or obligation so long as you agree to
// the following:
//
//  1.  You must indemnify Lowell Boggs against any and all financial
//      obligations caused by its use, misuse, function, or malfunction.
//      Further, you acknowledge that there is no warranty of any kind,
//      whatsoever.
//
//  2.  You agree not to attempt to patent any portion of this original
//      work -- though you may attempt to patent your own extensions to
//      it if you so choose.
//
//  3.  You keep this copyright notice with the file and all copies
//      of the file and do not change it anyway except language translation.
//
// You are responsible for enforcing your own compliance with these
// conditions and may not use this source file if you cannot agree to the
// above terms and conditions.
//
// Warning:  not all files in this directory structure are covered by the
// same copyright.  Some of them are part of the GNU source distribution
// and you must obey the GPL copyright for those files.



#include "glob.h"
#include "directory.h"
#include "fnmatch.h"
#include <iostream>
#include <sys/types.h>
#include <sys/stat.h>

#ifndef _S_IFDIR
#ifdef S_IFDIR
#define _S_IFDIR S_IFDIR
#endif
#endif



using namespace std;

static inline bool isDotDirReference(char const *p)
  // Return true if the specified file name is either "." or ".."
{
  if(   (p[0] == '.' )
     && (   (p[1] == 0)
         || ( (p[1] == '.' ) && (p[2] == 0) )
        )
    )
  {
    return true;
  }

  return false;

}

void glob(string const &pattern, list<string> &output)
{
  list<string> rv;

  if(pattern.size() < 1)
  {
    output.swap(rv);
    return;
  }


  ////////////////////////////////////////////////////////////////////////
  // step 1 -- split the pattern up into a list of directory node names

  list<string> nodes;

 // step 1a -- handle the special condition at the start of a pathname -- if the first node
 // is / or c:/ then we want to keep the / in the name, otherwise its a delimiter between
 // nodes in the pathname
  {
    string::const_iterator first = pattern.begin(), last = pattern.end();
    string node;
  
    char c = *first;
  
    if(c == '/' || c == '\\')
    {
      node += c;
  
      ++first;

      if(first != last && (*first == '/' || *first == '\\') )
      {
        node += '/';
        ++first;
      }
  
      while(first != last && (*first == '/' || *first == '\\') )
      {
        ++first;   // eat extranous slashes
      }
  
    }
    else
    if( (c >= 'a' && c <='z') || (c >= 'A' && c <= 'Z' ) )
    {
      node += c;
  
      ++first;
  
      if( (first != last) && ( *first == ':' ) )
      {
        // we are dealing with a dos style c:/file stuff
  
        node += ':';
  
        ++first;

        if(first != last && (*first == '/' || *first == '\\') )
        {
          node += '/';
  
          ++first;
        }

  
        while(first != last && (*first == '/' || *first == '\\') )
        {
          ++first;   // eat extraneous slashes
        }
  
      }
  
    }
  
    // while loop to incorporate the current node into node name list
  
    while(first != last)
    {
      c = *first;
  
      if(c == '/' ||c == '\\')
      {
        // end of current node
  
        nodes.push_back(node);
        node.resize(0);
  
        while(first != last && (*first == '/' || *first == '\\'))
          ++first;

        if(first != last)
          node += *first;
        else
          break;
  
      }
      else
        node += c;

      ++first;
    }
  
    if(node.size())
      nodes.push_back(node);
  
  }

  ////////////////////////////////////////////////////////////////////////
  // step 2 build up and filter down file names based on the nodes list and patterns
  //        therein.  As we process each node, we assume that the return value list has
  //        items in it which match part of the pattern and at each node will filter or
  //        more items.  When done processing the node we swap the new items with the current
  //        rv stuff and delete the old rv stuff.

 {

   list<string>::const_iterator first, last;

   list<string> candidates;

   for(first = nodes.begin(), last = nodes.end(); first != last; ++first)
   {
      string const &node = *first;

      if(node.size() == 0)
      {
        cerr << __FILE__ << ':' << __LINE__ << ":  BUG parsing pathname '" << pattern << "'" << endl;

        output.swap(rv);  // this shouldn't be possible
        return;
      }

      bool hasPattern = false;

      {
        // see if this node has file name pattern matching characters

        string::const_iterator begin = node.begin(), end = node.end();

        while(begin != end)
        {
          if(*begin == '\\')
          {
            ++begin;
            if(begin != end)
              ++begin;
          }
          else
          if( *begin == '[' || *begin == '*' || *begin == '?' )
          {
            if(*begin == '[')
            {
              // we have an open charset, check to see if the closer is found as well

              string::const_iterator scan = begin;

              ++scan;

              while(scan != end)
              {
                if(*scan == '\\')
                {
                  ++scan;
                  if(scan != end)
                    ++scan;
                }
                else
                {
                  if(*scan == ']')
                  {
                    hasPattern = true;
                    break;
                  }

                  ++scan;
                }
              }

            }
            else
              hasPattern = true;
            break;
          }
          else
            ++begin;
        }
      }

      if(hasPattern)
      {
        // this node has a file name pattern

        if(rv.size() == 0)
        {
          // ok the first level of access has patterns
          // read the top level directory and populate it with matching files

          if(node.size() == 0)
            break;

          bool isDosRoot = ( node.size() > 2 && node[1] == ':' );

          if(node[0] == '/' || isDosRoot )
          {
            // we have a root directory envolved

            string pattern;

            string header="/";

            if(isDosRoot)
            {
              pattern = node.substr(3, node.size()-3);
              header  = node.substr(0,3);
            }
            else
              pattern = node.substr(1, node.size()-1);


            DIR *rootDir = opendir(header.c_str());

            if(!rootDir)
              break;

            struct dirent *p;

            while( (p = readdir(rootDir)) )
            {
              if( isDotDirReference(p->d_name))
                continue;

              if(0 == fnmatch((char*)(pattern.c_str()), p->d_name, FNM_PATHNAME))
                candidates.push_back(header + p->d_name);
                
            }

            closedir(rootDir);

          }
          else
          {
            // we have a subdir

            DIR *rootDir = opendir(".");

            if(!rootDir)
              break;

            struct dirent *p;


            while( (p = readdir(rootDir)) )
            {
              if( isDotDirReference(p->d_name))
                continue;

              if(0 == fnmatch(  (char*)(node.c_str()), p->d_name, FNM_PATHNAME))
                candidates.push_back(p->d_name);
                
            }

            closedir(rootDir);
          }



        }
        else
        {
          // since we already have stuff in the rv, it can't be the root dir we are
          // looking at, go through the list of extant directories and see if the
          // any of them has a child matching the current node.  Any file in the extant
          // list in rv that isn't a directory, discard

          list<string>::const_iterator begin, end;

          // there are already multiple strings in the return value
          // To each one, add the current node as a trailing directory
          // entry, then check to see if said file exists.  If so, put it in the
          // candidates list -- if not, pitch it.

          for(begin = rv.begin(), end = rv.end(); begin != end; ++begin)
          {
            struct stat buf;

            string const &beg = *begin;

            if(0 == stat(beg.c_str(), &buf) )
            {
              // the current thing is a file or directory

              if(buf.st_mode &  _S_IFDIR)
              {
                // at this point we are only interested in directories

                DIR *dir = opendir(beg.c_str());

                if(dir)
                {
                  struct dirent *p;

                  while( (p = readdir(dir)) )
                  {
                    if( isDotDirReference(p->d_name))
                      continue;

                    if(0 == fnmatch( (char*)(node.c_str()), p->d_name, FNM_PATHNAME))
                    {
                      candidates.push_back(beg + "/" + p->d_name);
                    }
                  }


                  closedir(dir);
                }

              }

            }


          }


        }


      }
      else
      {
        // this node has only constant text -- just append it to the end of all extant
        // candidates

        if(rv.size() == 0)
        {
          // no current candidates -- so make this be the first candidate

          candidates.push_back(node);

        }
        else
        {
          list<string>::const_iterator begin, end;

          // there are already multiple strings in the return value
          // To each one, add the current node as a trailing directory
          // entry, then check to see if said file exists.  If so, put it in the
          // candidates list -- if not, pitch it.

          for(begin = rv.begin(), end = rv.end(); begin != end; ++begin)
          {
            string const &beg = *begin;
            string work = beg + "/" + node;

            if(access(work.c_str(), F_OK) == 0)
            {
              // this file or directory actually exists...

              candidates.push_back(work);

            }


          }
        }
      }


     // now swap the candidates with the rv:  this destructs the old rv and leaves the candidates
     // as the new rv.

     rv.swap(candidates);

     candidates.resize(0);

     if(rv.size() == 0)
     {
       break; // nothing more to do
     }

   }


 }

  output.swap(rv);  

}

